# Transformer embeddings
#
# Author: Giovanni Spadaro - https://giovannispadaro.it
# Project: https://github.com/Giovo17/tfs-mt
# Documentation: https://giovo17.github.io/tfs-mt
#
# Copyright (c) Giovanni Spadaro.
# All rights reserved.
#
# This source code is licensed under the license found in the LICENSE file in the root directory of this source tree.

import math
import os
import typing

import torch
import torch.nn as nn
from jaxtyping import Float

from .data_utils import download_glove


class TokenizerNotSuppliedError(Exception):
    def __init__(
        self,
        msg="Tokenizer not supplied. When loading pretrained Glove embeddings the tokenizer has to be supplied in order to map GloVe words to vocab entries.",
    ):
        super().__init__(msg)


class VocabNotBuiltError(Exception):
    def __init__(self, msg="Tokenizer vocabulary not built."):
        super().__init__(msg)


class EmbeddingDimError(Exception):
    def __init__(self, d_model, from_pretrained):
        msg = f"d_model cannot be None while from_pretrained is False, got d_model = {d_model} and from_pretrained = {from_pretrained}."
        super().__init__(msg)


class EmbeddingTypePathError(Exception):
    def __init__(self, from_pretrained, pretrained_emb_type, pretrained_emb_path):
        msg = f"pretrained_emb_type and pretrained_emb_path cannot be None while from_pretrained is true, \
                got from_pretrained = {from_pretrained}, pretrained_emb_type = {pretrained_emb_type} and pretrained_emb_path = {pretrained_emb_path}."
        super().__init__(msg)


class EmbeddingTypeNotImplementedError(Exception):
    def __init__(self, emb_type):
        msg = f"Embedding type not implemented, got emb_type = {emb_type}"
        super().__init__(msg)


class IncompatibleEmbeddingsDimError(Exception):
    def __init__(self, token_embeddings_shape):
        msg = f"Expected (batch size, max token sequence length, model dimension) got {token_embeddings_shape}"
        super().__init__(msg)


class Embedding(nn.Module):
    """Transformer embeddings layer.

    This implementation uses a randomly initialized embedding lookup table with dimension `[vocab_size, d_model]`.

    Note: There's the possibility of loading pretrained embeddings from GloVe.
        This choice has been made to achieve acceptable performances with low resources training and limited time training.

    Note:
        GloVe embeddings available for English only.

    Args:
        vocab_size (int): Number of tokens in vocabulary.
        d_model (int | None, optional): Model dimension. Defaults to None.
        from_pretrained (bool, optional): Load embeddings from pretrained. Defaults to False.
        pretrained_emb_type (str | None, optional): Type of pretrained embeddings. Defaults to None.
        pretrained_emb_path (str | None, optional): Path of pretrained embeddings. Defaults to None.

    Raises:
        EmbeddingDimError: Raised when the provided embedding dimension does not match the expected size.
        EmbeddingTypePathError: Raised when the embedding type or file path is invalid or not found.
        TokenizerNotSuppliedError: Raised when a tokenizer is required but has not been supplied.
        VocabNotBuiltError: Raised when an operation requiring a built vocabulary is attempted before the vocabulary is constructed.
    """

    def __init__(
        self,
        vocab_size: int,
        d_model: int | None = None,
        from_pretrained: bool = False,
        pretrained_emb_type: str | None = None,
        pretrained_emb_path: str | None = None,
        **kwargs,
    ):
        super().__init__()

        if d_model is None and not from_pretrained:
            raise EmbeddingDimError(d_model, from_pretrained)

        if from_pretrained:
            if d_model is not None:
                print(f"Ignoring d_model ({d_model}). The embeddings dim will be inferred from pretrained.")
            if pretrained_emb_type is None or pretrained_emb_path is None:
                raise EmbeddingTypePathError(from_pretrained, pretrained_emb_type, pretrained_emb_path)
            if pretrained_emb_type == "GloVe" and "tokenizer" not in kwargs:
                raise TokenizerNotSuppliedError()
            if kwargs["tokenizer"].vocab_size == 0:
                raise VocabNotBuiltError()

            if pretrained_emb_type == "GloVe":
                embeddings_dim, embeddings_lut = self.load_pretrained(
                    pretrained_emb_path, pretrained_emb_type, tokenizer=kwargs["tokenizer"]
                )
            else:
                embeddings_dim, embeddings_lut = self.load_pretrained(pretrained_emb_path, pretrained_emb_type)

            # The following operations consist in rescaling the norm of GloVe pretrained embeddings.
            # This is done to have source and target embeddings with a comparable norm to stabilize training.
            embeddings_lut_dummy = nn.Embedding(embeddings_lut.weight.shape[0], embeddings_lut.weight.shape[1])
            nn.init.xavier_uniform_(embeddings_lut_dummy.weight)

            embeddings_lut_dummy_norm = torch.mean(embeddings_lut_dummy.weight.data.norm(dim=1))
            embeddings_lut_norm = torch.mean(embeddings_lut.weight.data.norm(dim=1))

            rescale_coeff = embeddings_lut_dummy_norm / embeddings_lut_norm

            embeddings_lut.weight.data = embeddings_lut.weight.data * rescale_coeff

        else:
            if d_model is None:
                raise EmbeddingDimError(d_model, from_pretrained)
            embeddings_dim = d_model
            # Randomly initialized lookup table.
            embeddings_lut = nn.Embedding(vocab_size, embeddings_dim)
            nn.init.xavier_uniform_(embeddings_lut.weight)

        self.d_model = embeddings_dim
        self.scaling_factor = math.sqrt(self.d_model)
        self.embeddings_lut = embeddings_lut

    def load_pretrained(self, embeddings_path: str, emb_type: str = "GloVe", **kwargs) -> tuple[int, nn.Embedding]:
        """Loads pretrained GloVe embedding into the embedding lookup table."""
        if emb_type == "GloVe":
            tokenizer = kwargs["tokenizer"]

            if not os.path.isfile(embeddings_path):
                output_dir = "/".join(embeddings_path.split("/")[:-2])
                download_glove(output_dir, glove_version=embeddings_path.split("/")[-2])

            with open(embeddings_path, encoding="utf-8") as f:
                embeddings_dim = len(f.readline().strip().split()) - 1
            embeddings_lut = nn.Embedding(tokenizer.vocab_size, embeddings_dim)

            # NOTE The vocab extension with GloVe tokens is handled by the tokenizer.
            # Here GloVe token embeddings are mapped to the corresponding entry in the embeddings lookup table
            with open(embeddings_path, encoding="utf-8") as f:
                for line in f:
                    parts = line.strip().split()
                    idx, _ = tokenizer.encode(parts[0])
                    # If tokenization if non perfectly compatible with the GloVe one, idx can be a sequence of tokens longer then 3
                    # (Consider the first and last tokens coming out of tokenizer.encode are SOS_TOKEN and EOS_TOKEN)
                    # This is a useful check to avoid overwriting in the embeddings_lut matrix
                    if len(idx) > 3:
                        continue
                    idx = idx[1]  # The first token coming out of tokenizer.encode is SOS_TOKEN
                    if len(parts[1:]) != embeddings_dim:  # Skip unhandled tokens with spaces, eg. "1 3/4"
                        continue
                    try:
                        token_emb = torch.tensor([float(x) for x in parts[1:]], dtype=torch.float32)
                    except ValueError:
                        continue
                    else:
                        embeddings_lut.weight.data[idx].copy_(token_emb)
        else:
            raise EmbeddingTypeNotImplementedError(emb_type)

        return embeddings_dim, embeddings_lut

    def forward(self, token_ids: Float[torch.Tensor, "B S"]) -> Float[torch.Tensor, "B S D"]:
        """Get token embeddings.

        Args:
            token_ids (Float[torch.Tensor, "B S"]): Input batch of token_ids. Where B is the batch size and S is the sequence length.

        Returns:
            Float[torch.Tensor, "B S D"]: Output batch of token embeddings. Where D is d_model.
        """

        embeddings = self.embeddings_lut(token_ids)

        # In the embedding layers, we multiply those weights by sqrt(d_model) (Attention is all you need page 5)
        embeddings = embeddings * self.scaling_factor

        return embeddings


class SinusoidalPositionalEncoding(nn.Module):
    """Sinusoidal Positional Encoding implementation from the [original paper](https://arxiv.org/abs/1706.03762).

    $$
    \\begin{align*}
    PE_{(pos,2i)} &= \\sin(pos/10000^{2i/d_{model}}) \\\\
    PE_{(pos,2i+1)} &= \\cos(pos/10000^{2i/d_{model}})
    \\end{align*}
    $$

    Args:
        d_model (int): Model dimension.
        dropout_prob (float, optional): Dropout probability. Defaults to 0.1.
        max_sequence_length (int, optional): Max sequence length. Defaults to 128.
    """

    def __init__(self, d_model: int, dropout_prob: float = 0.1, max_sequence_length: int = 128):
        super().__init__()

        # Vector of all possible positions in the sequence [max_sequence_length, 1]
        position_id = torch.arange(0, max_sequence_length).unsqueeze(1)
        i_idx = torch.arange(0, d_model, 2, dtype=torch.float32) / d_model  # 2i / d_model
        freq_vec = torch.pow(10000.0, -i_idx)  # 1 / (10000^(2i/d_model))

        pe_lut = torch.zeros(max_sequence_length, d_model)  # Init positional encoding lookup table
        pe_lut[:, 0::2] = torch.sin(position_id * freq_vec)  # Assign sine on even positions
        pe_lut[:, 1::2] = torch.cos(position_id * freq_vec)  # Assing cosine on odd positions

        # Registering this weights as buffers so that they will be saved in model state_dict,
        # but they won't appear in model.parameters so that optimizer will not change them
        self.register_buffer("pe_lut", pe_lut)

        self.dropout = nn.Dropout(dropout_prob)

    def forward(self, token_embeddings: Float[torch.Tensor, "B S D"]) -> Float[torch.Tensor, "B S D"]:
        """Get token embeddings with positional information.

        Args:
            token_embeddings (Float[torch.Tensor, "B S D"]): Input batch of token embeddings.

        Raises:
            IncompatibleEmbeddingsDimError: Raised when the input embedding dimension is invalid.

        Returns:
            Float[torch.Tensor, "B S D"]: Token embeddings with added positional information.
        """
        pe_lut_tensor = typing.cast(torch.Tensor, self.pe_lut)
        if token_embeddings.ndim != 3 or token_embeddings.size(-1) != pe_lut_tensor.shape[1]:
            raise IncompatibleEmbeddingsDimError(token_embeddings.shape)

        positional_encodings = pe_lut_tensor[: token_embeddings.size(1)]

        # we apply dropout to the sums of the embeddings and the positional encodings in both the encoder and decoder stacks (Attention is all you need page 8)
        final_embedding = self.dropout(token_embeddings + positional_encodings)

        return final_embedding



# Ignite common utils
#
# This file is derived from PyTorch Ignite (BSD 3-Clause License)
# Copyright (c) 2018-present, PyTorch-Ignite team
# License for original code: BSD 3-Clause
#
# Modifications:
# Author: Giovanni Spadaro - https://giovannispadaro.it
# Project: https://github.com/Giovo17/tfs-mt
# Documentation: https://giovo17.github.io/tfs-mt
#
# Copyright (c) Giovanni Spadaro.
# License for modifications: MIT
#
# The full text of both licenses can be found in the LICENSE file in the root directory of this source tree.


from typing import Any

from ignite.contrib.engines import common
from ignite.engine import Engine
from torch.optim.optimizer import Optimizer

from .trackio_logger import TrackioLogger
from .wandb_logger import WandBLogger


def setup_wandb_logging(
    trainer: Engine,
    optimizers: Optimizer | dict[str, Optimizer] | None = None,
    evaluators: Engine | dict[str, Engine] | None = None,
    log_every_iters: int = 100,
    **kwargs: Any,
) -> WandBLogger:
    logger = WandBLogger(**kwargs)
    common._setup_logging(logger, trainer, optimizers, evaluators, log_every_iters)
    return logger


def setup_trackio_logging(
    trainer: Engine,
    optimizers: Optimizer | dict[str, Optimizer] | None = None,
    evaluators: Engine | dict[str, Engine] | None = None,
    log_every_iters: int = 100,
    **kwargs: Any,
) -> TrackioLogger:
    logger = TrackioLogger(**kwargs)
    common._setup_logging(logger, trainer, optimizers, evaluators, log_every_iters)
    return logger



# Visualizations script
#
# Author: Giovanni Spadaro - https://giovannispadaro.it
# Project: https://github.com/Giovo17/tfs-mt
# Documentation: https://giovo17.github.io/tfs-mt
#
# Copyright (c) Giovanni Spadaro.
# All rights reserved.
#
# This source code is licensed under the license found in the LICENSE file in the root directory of this source tree.

import os
import typing

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
import torch
import torch.nn as nn
import torch.nn.functional as F
from matplotlib.lines import Line2D
from omegaconf import OmegaConf
from torch.nn import CrossEntropyLoss
from torch.optim import AdamW
from torch.optim.lr_scheduler import LambdaLR

from tfs_mt.architecture import build_model
from tfs_mt.data_utils import WordTokenizer
from tfs_mt.embeddings import SinusoidalPositionalEncoding
from tfs_mt.training_utils import KLDivLabelSmoothingLoss, setup_lr_lambda_fn


def load_model_and_tokenizers():
    base_url = "https://huggingface.co/giovo17/tfs-mt/resolve/main/"
    src_tokenizer = WordTokenizer.from_pretrained(base_url + "src_tokenizer_word.json")
    tgt_tokenizer = WordTokenizer.from_pretrained(base_url + "tgt_tokenizer_word.json")

    model = build_model(
        config=base_url + "config-lock.yaml",
        from_pretrained=True,
        model_path=base_url + "model.safetensors",
    )
    model.eval()

    return model, src_tokenizer, tgt_tokenizer


def display_attention_matrix(
    selected_layer: int = -1, selected_attention_head: int = 0, pad_len: int = 9, annot_threshold: float = 0.05
):
    model, src_tokenizer, _ = load_model_and_tokenizers()

    example_en_string = "The Transformer architecture was introduced in 2017"

    tokens = src_tokenizer.tokenize(example_en_string)

    token_ids, mask = src_tokenizer.encode(tokens, pad_to_len=pad_len)

    # Reconstruct display tokens to include the padding tokens.
    # NOTE pad_to_len in tokenizer.encode() doesn't consider the SOS and EOS tokens
    display_tokens = tokens + ["<pad>"] * (pad_len - len(tokens) + 2)

    # Run the encode method just to populate attention weights in the encoder
    src_sequence = torch.tensor(token_ids, dtype=torch.long).unsqueeze(0)
    src_mask = torch.tensor(mask, dtype=torch.bool).unsqueeze(0)
    with torch.inference_mode():
        _ = model.encode(src_sequence, src_mask)

    # Select the attention weights of the last encoder layer
    attn_weights = model.encoder[selected_layer].self_attention.attn_weights
    # Select the head and the first and only element in the batch
    attn_weights = attn_weights[0, selected_attention_head, :, :]
    # The last rows of the attention matrix associated with padding results in nan values.
    # Replacing them with zeros for better visualization
    attn_weights = torch.where(torch.isnan(attn_weights), torch.zeros_like(attn_weights), attn_weights)

    # Attention mask has shape [B, 1, 1, S] cause the mask is the across every head
    attn_mask = model.attention_mask[0, 0, :, :]

    attn_weights = pd.DataFrame(attn_weights.detach().numpy())

    # Display only values higher than annot_threshold for better visualization
    annot = attn_weights.map(lambda v: f"{v:.2f}" if v > annot_threshold else "")

    _, ax = plt.subplots(figsize=(10, 8))
    sns.heatmap(
        attn_weights,
        cmap="viridis",
        annot=annot,
        square=True,
        vmin=0,
        vmax=1,
        fmt="",
        xticklabels=display_tokens,
        yticklabels=display_tokens,
    )

    # Add red borders where mask is True
    for i in range(attn_mask.shape[0]):
        for j in range(attn_mask.shape[1]):
            if not attn_mask[i, j]:  # The mask values are true where attention has to be computed
                ax.plot([j, j + 1], [i, i + 1], color="red", linewidth=1)
                ax.plot([j + 1, j], [i, i + 1], color="red", linewidth=1)

    plt.xticks(rotation=45, ha="right")
    plt.yticks(rotation=0)
    plt.tight_layout()
    plt.savefig(os.path.join(os.getcwd(), "docs/assets/img/attention_matrix.png"))
    plt.show()


def display_causal_attention_matrix(
    selected_layer: int = -1, selected_attention_head: int = 0, pad_len: int = 9, annot_threshold: float = 0.05
):
    model, src_tokenizer, tgt_tokenizer = load_model_and_tokenizers()

    example_en_string = "The Transformer architecture was introduced in 2017"
    example_it_string = "L'architettura Transformer è stata introdotta nel 2017"

    src_tokens = src_tokenizer.tokenize(example_en_string)
    src_token_ids, src_mask = src_tokenizer.encode(src_tokens, pad_to_len=pad_len)

    tgt_tokens = tgt_tokenizer.tokenize(example_it_string)

    tgt_token_ids, tgt_mask = tgt_tokenizer.encode(tgt_tokens, pad_to_len=pad_len)  # pad_to_len=len(tgt_tokens_list)+2)

    display_tokens = tgt_tokens + ["<pad>"] * (pad_len - len(tgt_tokens) + 2)

    # Run the decode method just to populate attention weights in the encoder
    src_sequence = torch.tensor(src_token_ids, dtype=torch.long).unsqueeze(0)
    src_mask_tensor = torch.tensor(src_mask, dtype=torch.bool).unsqueeze(0)
    tgt_sequence = torch.tensor(tgt_token_ids, dtype=torch.long).unsqueeze(0)
    tgt_mask_tensor = torch.tensor(tgt_mask, dtype=torch.bool).unsqueeze(0)
    with torch.inference_mode():
        encoder_representation = model.encode(src_sequence, src_mask_tensor)
        _ = model.decode(tgt_sequence, encoder_representation, tgt_mask_tensor, src_mask_tensor)

    # Select the attention weights of the last decoder layer
    attn_weights = model.decoder[selected_layer].self_attention.attn_weights
    # Select the head and the first and only element in the batch
    attn_weights = attn_weights[0, selected_attention_head, :, :]
    attn_weights = torch.where(torch.isnan(attn_weights), torch.zeros_like(attn_weights), attn_weights)

    # Attention mask has shape [B, 1, 1, S] cause the mask is the across every head
    causal_attn_mask = model.causal_attention_mask[0, 0, :, :]

    attn_weights = pd.DataFrame(attn_weights.detach().numpy())

    annot = attn_weights.map(lambda v: f"{v:.2f}" if v > annot_threshold else "")

    _, ax = plt.subplots(figsize=(10, 8))
    sns.heatmap(
        attn_weights,
        cmap="viridis",
        annot=annot,
        square=True,
        vmin=0,
        vmax=1,
        fmt="",
        xticklabels=display_tokens,
        yticklabels=display_tokens,
    )

    # Add red borders where mask is True
    for i in range(causal_attn_mask.shape[0]):
        for j in range(causal_attn_mask.shape[1]):
            if not causal_attn_mask[i, j]:  # The mask values are true where attention has to be computed
                ax.plot([j, j + 1], [i, i + 1], color="red", linewidth=1)
                ax.plot([j + 1, j], [i, i + 1], color="red", linewidth=1)

    plt.xticks(rotation=45, ha="right")
    plt.yticks(rotation=0)
    plt.tight_layout()
    plt.savefig(os.path.join(os.getcwd(), "docs/assets/img/causal_attention_matrix.png"))
    plt.show()


def display_cross_attention_matrix(
    selected_layer: int = -1,
    selected_attention_head: int = 0,
    src_text: str = "The Transformer architecture was introduced in 2017",
    tgt_text: str = "L'architettura Transformer è stata introdotta nel 2017",
    pad_len: int = 9,
    annot_threshold: float = 0.05,
    save_path: str = "cross_attention_matrix.png",
):
    model, src_tokenizer, tgt_tokenizer = load_model_and_tokenizers()

    src_tokens = src_tokenizer.tokenize(src_text)
    src_token_ids, src_mask = src_tokenizer.encode(src_tokens, pad_to_len=pad_len)

    tgt_tokens = tgt_tokenizer.tokenize(tgt_text)

    tgt_token_ids, tgt_mask = tgt_tokenizer.encode(tgt_tokens, pad_to_len=pad_len)

    src_display_tokens = src_tokens + ["<pad>"] * (pad_len - len(src_tokens) + 2)
    tgt_display_tokens = tgt_tokens + ["<pad>"] * (pad_len - len(tgt_tokens) + 2)

    # Run the decode method just to populate attention weights in the encoder
    src_sequence = torch.tensor(src_token_ids, dtype=torch.long).unsqueeze(0)
    src_mask_tensor = torch.tensor(src_mask, dtype=torch.bool).unsqueeze(0)
    tgt_sequence = torch.tensor(tgt_token_ids, dtype=torch.long).unsqueeze(0)
    tgt_mask_tensor = torch.tensor(tgt_mask, dtype=torch.bool).unsqueeze(0)
    with torch.inference_mode():
        encoder_representation = model.encode(src_sequence, src_mask_tensor)
        _ = model.decode(tgt_sequence, encoder_representation, tgt_mask_tensor, src_mask_tensor)

    # Select the attention weights of the last decoder layer
    attn_weights = model.decoder[selected_layer].cross_attention.attn_weights
    # Select the head and the first and only element in the batch
    attn_weights = attn_weights[0, selected_attention_head, :, :]
    attn_weights = torch.where(torch.isnan(attn_weights), torch.zeros_like(attn_weights), attn_weights)

    # Attention mask has shape [B, 1, 1, S] cause the mask is the across every head
    cross_attn_mask = model.cross_attention_mask[0, 0, :, :]

    attn_weights = pd.DataFrame(attn_weights.detach().numpy())

    annot = attn_weights.map(lambda v: f"{v:.2f}" if v > annot_threshold else "")

    _, ax = plt.subplots(figsize=(10, 8))
    sns.heatmap(
        attn_weights,
        cmap="viridis",
        annot=annot,
        square=True,
        vmin=0,
        vmax=1,
        fmt="",
        xticklabels=src_display_tokens,
        yticklabels=tgt_display_tokens,
    )

    # Add red borders where mask is True
    for i in range(cross_attn_mask.shape[0]):
        for j in range(cross_attn_mask.shape[1]):
            if not cross_attn_mask[i, j]:  # The mask values are true where attention has to be computed
                ax.plot([j, j + 1], [i, i + 1], color="red", linewidth=1)
                ax.plot([j + 1, j], [i, i + 1], color="red", linewidth=1)

    plt.xticks(rotation=45, ha="right")
    plt.yticks(rotation=0)
    plt.tight_layout()
    plt.savefig(os.path.join(os.getcwd(), "docs/assets/img/", save_path))
    plt.show()


def display_positional_encoding():
    pos_encoder = SinusoidalPositionalEncoding(
        d_model=16,
        max_sequence_length=500,
    )

    pe_lut: torch.Tensor = pos_encoder.pe_lut
    pe = pe_lut.numpy().transpose()

    plt.figure(figsize=(10, 6))
    sns.heatmap(pe, cmap="RdBu", center=0, annot=False, yticklabels=True, xticklabels=100)
    plt.xlabel("Sequence position")
    plt.ylabel("Embedding dimension")
    plt.tight_layout()
    plt.savefig(os.path.join(os.getcwd(), "docs/assets/img/positional_encoding.png"))
    plt.show()


def display_sinusoidal_signals():
    fig, axes = plt.subplots(8, 1, figsize=(10, 6), sharex=True)

    t = np.linspace(0, 1024, num=1024)

    pos_encoder = SinusoidalPositionalEncoding(
        d_model=8,
        max_sequence_length=1024,
    )

    x_pin_index = 407

    for _, (ax, i) in enumerate(zip(axes, range(8), strict=False)):
        pe_lut_tensor = typing.cast(torch.Tensor, pos_encoder.pe_lut)
        signal = pe_lut_tensor[:, i]

        ax.plot(t, signal)

        y_label = round(signal.numpy()[x_pin_index], 2)
        ax.set_ylabel(y_label, rotation=0, labelpad=20, fontsize=20, color="black", weight="bold")
        ax.yaxis.set_label_coords(-0.05, 0.10)

        ax.set_yticks([])

        for spine in ax.spines.values():
            spine.set_linewidth(0)

        ax.grid(False)

    # Add vertical line to show the values of this pseudo positional embedding at example index
    for ax in axes:
        ax.axvline(x=x_pin_index, color="#000000", linewidth=3, linestyle="-", alpha=0.8)

    plt.xlim(0, 1024)
    plt.tight_layout()
    plt.savefig(os.path.join(os.getcwd(), "docs/assets/img/sinusoidal_signals.png"))
    plt.show()


def display_square_wave_signals():
    fig, axes = plt.subplots(8, 1, figsize=(10, 6), sharex=True)

    periods = [8, 16, 32, 64, 128, 256, 512, 1024]
    t = np.linspace(0, 1024, num=1024)
    x_pin_index = 407

    for _, (ax, period) in enumerate(zip(axes, periods, strict=False)):
        # Square wave signal
        signal = 1 - (np.mod(t, period) < period / 2).astype(int)

        ax.step(t, signal, where="post", linewidth=1.5)
        ax.fill_between(t, 0, signal, step="post", alpha=0.1)

        ax.set_ylabel(signal[x_pin_index], rotation=0, labelpad=20, fontsize=20, color="black", weight="bold")
        ax.yaxis.set_label_coords(-0.05, 0.10)

        ax.set_yticks([])

        for spine in ax.spines.values():
            spine.set_linewidth(0)

        ax.grid(False)

    # Add vertical line to show the values of this pseudo positional embedding at example index
    for ax in axes:
        ax.axvline(x=x_pin_index, color="#000000", linewidth=3, linestyle="-", alpha=0.8)

    plt.xlim(0, 1024)
    plt.tight_layout()
    plt.savefig(os.path.join(os.getcwd(), "docs/assets/img/square_wave_signals.png"))
    plt.show()


def display_lr_schedule_wsd(config):
    config.training_hp.lr_scheduler.type = "wsd"

    batch_size = 16
    input_dim = 10

    def get_batch():
        x = torch.randn(batch_size, input_dim)
        y = torch.randint(0, 2, (batch_size,)).long()
        return x, y

    # Dummy model
    model = nn.Linear(input_dim, 2)
    criterion = CrossEntropyLoss()
    optimizer = AdamW(model.parameters(), lr=1)

    lr_lambda = setup_lr_lambda_fn(config)

    scheduler = LambdaLR(optimizer, lr_lambda=lr_lambda)

    lrs = []
    global_step = 0
    for _epoch in range(config.training_hp.num_epochs):
        for _ in range(config.num_train_iters_per_epoch):
            x, y = get_batch()
            optimizer.zero_grad()
            logits = model(x)
            loss = criterion(logits, y)
            loss.backward()
            optimizer.step()

            lrs.append(optimizer.param_groups[0]["lr"])

            scheduler.step()
            global_step += 1

    plt.figure(figsize=(10, 5))
    sns.lineplot(x=list(range(len(lrs))), y=lrs)
    plt.xlabel("Step")
    plt.ylabel("Learning rate")
    plt.tight_layout()
    plt.savefig(os.path.join(os.getcwd(), f"docs/assets/img/lr_scheduling_{config.training_hp.lr_scheduler.type}.png"))
    plt.show()


def display_lr_schedule_original(config):
    config.training_hp.lr_scheduler.type = "original"

    batch_size = 16
    input_dim = 5

    def get_batch():
        x = torch.randn(batch_size, input_dim)
        y = torch.randint(0, 2, (batch_size,)).long()
        return x, y

    # Dummy model
    model = nn.Linear(input_dim, 2)
    criterion = CrossEntropyLoss()
    optimizer = AdamW(model.parameters(), lr=1)

    plt.figure(figsize=(10, 5))

    graph_configs = [(100, 10000), (256, 10000), (256, 15000)]
    for d_model, warmup_iters in graph_configs:
        config.model_configs[config.chosen_model_size].d_model = d_model
        config.training_hp.lr_scheduler.warmup_iters = warmup_iters

        lr_lambda = setup_lr_lambda_fn(config)
        scheduler = LambdaLR(optimizer, lr_lambda=lr_lambda)

        lrs = []
        global_step = 0
        for _epoch in range(config.training_hp.num_epochs):
            for _ in range(config.num_train_iters_per_epoch):
                x, y = get_batch()
                optimizer.zero_grad()
                logits = model(x)
                loss = criterion(logits, y)
                loss.backward()
                optimizer.step()

                lrs.append(optimizer.param_groups[0]["lr"])

                scheduler.step()
                global_step += 1

        sns.lineplot(x=list(range(len(lrs))), y=lrs, label=f"d_model={d_model}, warmup={warmup_iters}")

    plt.xlabel("Step")
    plt.ylabel("Learning rate")
    plt.legend()
    plt.tight_layout()
    plt.savefig(os.path.join(os.getcwd(), f"docs/assets/img/lr_scheduling_{config.training_hp.lr_scheduler.type}.png"))
    plt.show()


def display_label_smoothing(
    config, label_smoothing: float = 0.4, train_examples: int = 5, vocab_size: int = 8, pad_idx: int = 1
):
    criterion = KLDivLabelSmoothingLoss(vocab_size=vocab_size, padding_idx=pad_idx, smoothing=label_smoothing)

    outputs = F.softmax(torch.rand(train_examples, vocab_size), dim=1)

    pred = outputs.log_softmax(-1)

    # Make random target tensor
    non_pad_tokens = list(range(vocab_size))
    non_pad_tokens.remove(pad_idx)
    values = torch.tensor(non_pad_tokens)

    # Sample indices without replacement
    indices = torch.multinomial(torch.ones(len(values)), num_samples=train_examples - 1, replacement=False)
    target = values[indices]

    # Manually insert a pad token at -2 position to show the effect of padding
    target = torch.cat([target[:-1], torch.tensor([pad_idx]), target[-1:]])

    _, true_dist = criterion(pred, target, return_true_dist=True)

    LS_data = pd.concat([
        pd.DataFrame({
            "target distribution": true_dist[x, y].flatten(),
            "columns": y,
            "rows": x,
        })
        for y in range(vocab_size)
        for x in range(train_examples)
    ])

    heatmap_data = LS_data.pivot(index="rows", columns="columns", values="target distribution")

    plt.figure(figsize=(8, 5))
    sns.heatmap(heatmap_data, cmap="viridis", cbar_kws={"label": "target distribution"}, square=True, vmin=0, vmax=1)
    plt.xlabel(None)
    plt.ylabel(None)
    plt.tight_layout()
    plt.savefig(os.path.join(os.getcwd(), "docs/assets/img/label_smoothing.png"))
    plt.show()


def display_training_loss(config, csv_path):
    df = pd.read_csv(csv_path)

    plt.figure(figsize=(10, 6))

    # Smoothed loss with Exponential Moving Average
    df["smoothed_loss"] = df["training/train_loss"].ewm(span=20).mean()

    sns.lineplot(data=df, x="_step", y="training/train_loss", label="Loss", color="blue")
    # Smoothed loss
    sns.lineplot(data=df, x="_step", y="smoothed_loss", label="Loss (EMA)", color="blue", alpha=0.5)

    # Add vertical lines for epoch boundaries
    iters_per_epoch = config.num_train_iters_per_epoch
    max_step = df["_step"].max()
    for i in range(1, int(max_step // iters_per_epoch) + 1):
        plt.axvline(x=i * iters_per_epoch, color="red", linestyle="--", alpha=0.2)

    plt.xlabel("Step")
    plt.ylabel("Loss")
    plt.xlim(left=0)
    plt.ylim(bottom=0)
    plt.legend(loc="upper right")
    plt.grid(alpha=0.3)
    plt.tight_layout()
    plt.savefig(os.path.join(os.getcwd(), "docs/assets/img/training_loss.png"))
    plt.show()


def display_performance_metrics(csv_path):
    df = pd.read_csv(csv_path)

    metrics_map = {
        "test_eval/Bleu": "BLEU smooth 1",
        "test_eval/Bleu_smooth_2": "BLEU smooth 2",
        "test_eval/Rouge/Rouge-L-F": "ROUGE-L-F",
        "test_eval/Rouge/Rouge-2-F": "ROUGE-2-F",
        "test_eval/Perplexity": "Perplexity",
    }

    perplexity_col = "test_eval/Perplexity"
    primary_metrics = {k: v for k, v in metrics_map.items() if k != perplexity_col}

    fig, ax1 = plt.subplots(figsize=(12, 6))
    ax2 = ax1.twinx()  # Second y-axis sharing the same x-axis

    color_map = {
        "BLEU smooth 1": "tab:blue",
        "BLEU smooth 2": "tab:purple",
        "ROUGE-L-F": "tab:orange",
        "ROUGE-2-F": "tab:green",
        "Perplexity": "tab:red",
    }

    # Plot primary metrics (BLEU, ROUGE) on ax1
    for col, label in primary_metrics.items():
        if col in df.columns:
            data = df[["_step", col]].dropna(subset=[col])
            ax1.plot(data["_step"], data[col], label=label, marker="o", color=color_map[label])

    # Plot Perplexity on ax2
    if perplexity_col in df.columns:
        data = df[["_step", perplexity_col]].dropna(subset=[perplexity_col])
        ax2.plot(
            data["_step"],
            data[perplexity_col],
            label="Perplexity",
            marker="o",
            color=color_map["Perplexity"],
            linestyle="--",
        )

    ax1.set_xlabel("Step")
    ax1.set_ylabel("Score (BLEU / ROUGE)")
    ax2.set_ylabel("Score (Perplexity)", color=color_map["Perplexity"])
    ax2.tick_params(axis="y", labelcolor=color_map["Perplexity"])

    ax1.set_xlim(left=-2500)
    ax1.grid(alpha=0.3)

    custom_handles = [
        Line2D([0], [0], color=color_map[label], marker="o", label=label)
        for label in ["BLEU smooth 1", "BLEU smooth 2", "ROUGE-L-F", "ROUGE-2-F"]
    ] + [Line2D([0], [0], color=color_map["Perplexity"], marker="o", linestyle="--", label="Perplexity")]
    ax1.legend(handles=custom_handles, bbox_to_anchor=(1.15, 1), loc="upper left")

    plt.tight_layout()
    plt.savefig(os.path.join(os.getcwd(), "docs/assets/img/performance_metrics.png"))
    plt.show()


if __name__ == "__main__":
    config_path = os.path.join(os.getcwd(), "src/tfs_mt/configs/config.yml")
    config = OmegaConf.load(config_path)

    sns.set_context("talk", font_scale=0.95, rc={"font.size": 14})

    torch.manual_seed(42)

    os.makedirs(os.path.join(os.getcwd(), "docs/assets/img"), exist_ok=True)

    # Attention matrix ------------------------------------------------------------------------------------------------

    # config.chosen_model_size = "small"

    # selected_layer = -1
    # selected_attention_head = 1

    # display_attention_matrix(selected_layer, selected_attention_head)
    # display_causal_attention_matrix(selected_layer, selected_attention_head)
    # display_cross_attention_matrix(selected_layer, selected_attention_head)

    # # Positional Encoding ---------------------------------------------------------------------------------------------

    # display_positional_encoding()
    # display_sinusoidal_signals()
    # display_square_wave_signals()

    # # LR scheduler ----------------------------------------------------------------------------------------------------

    # config.num_train_iters_per_epoch = 28889
    # config.training_hp.num_epochs = 2

    # display_lr_schedule_original(config)
    # display_lr_schedule_wsd(config)

    # # Loss ------------------------------------------------------------------------------------------------------------

    # display_label_smoothing(config)

    # Experiments results graphs --------------------------------------------------------------------------------------

    config.num_train_iters_per_epoch = 28889

    results_csv = os.path.join(os.getcwd(), "data/eval/tfs_mt_small_260207-0915.csv")

    display_training_loss(config, results_csv)
    display_performance_metrics(results_csv)



# Transformer architecture
#
# Author: Giovanni Spadaro - https://giovannispadaro.it
# Project: https://github.com/Giovo17/tfs-mt
# Documentation: https://giovo17.github.io/tfs-mt
#
# Copyright (c) Giovanni Spadaro.
# All rights reserved.
#
# This source code is licensed under the license found in the LICENSE file in the root directory of this source tree.

import math
import os
import tempfile

import requests
import torch
import torch.nn as nn
from jaxtyping import Bool, Float
from omegaconf.dictconfig import DictConfig
from omegaconf.listconfig import ListConfig
from safetensors.torch import load_file

from .data_utils import BaseTokenizer
from .embeddings import Embedding, SinusoidalPositionalEncoding


class DimError(Exception):
    def __init__(self, d_model, num_heads):
        msg = f"d_model is not divisible by num_heads, got d_model = {d_model} and num_heads = {num_heads}."
        super().__init__(msg)


class MissingArgumentsError(Exception):
    def __init__(self, emb_type):
        msg_init = "Source" if emb_type == "src" else "Target"
        msg = f"{msg_init} embeddings initialization from pretrained has been requested, but {emb_type}_emb_pretrained_type or {emb_type}_emb_pretrained_path have not been provided."
        super().__init__(msg)


class MissingArgumentsGloVeError(Exception):
    def __init__(self, emb_type):
        msg_init = "Source" if emb_type == "src" else "Target"
        msg = f"{msg_init} embeddings initialization from GloVe pretrained has been requested, but {emb_type}_tokenizer has not been provided."
        super().__init__(msg)


class ModelSizeNotChoosen(Exception):
    def __init__(self, msg="Model size not choosen. Add chosen_model_size to config."):
        super().__init__(msg)


class InvalidFileExtensionException(Exception):
    def __init__(self, ext):
        msg = f"File extension not supported: `{ext}`. Please load a `.safetensors`, `.pt` or `.pth` file."
        super().__init__(msg)


class MultiHeadAttention(nn.Module):
    """MultiHead Attention for Transformer encoder and decoder. It handles both self and cross attention operations.

    Note: Implementation reference
        Following the implementation described in *Speech and Language Processing* by *Daniel Jurafsky* [[link](https://web.stanford.edu/~jurafsky/slp3/)].

    Args:
        d_model (int): Model dimension.
        num_heads (int): Number of attention heads.
        dropout_prob (float, optional): Dropout probability. Defaults to 0.1.
    """

    def __init__(self, d_model: int, num_heads: int, dropout_prob: float = 0.1):
        super().__init__()

        self.d_model = d_model
        self.num_heads = num_heads

        # NOTE Usually transformer models have d_model divisible by num_heads.
        # This guarantees that the attention heads' outputs are d_model shaped vectors when stacked together (considering each embedding vector)
        # In this implementation it has been preferred to remove this constraint. When exploiting the support of GloVe pretrained embeddings,
        # d_model is fixed to GloVe embeddings sizes, namely 25, 50, 100, 200, 300, so in this scenario num_heads would be limited to predefined set of values due to int quantization.
        # Considering the intermidiate output dimensions there will be no problems since the 3 initial projections matrices' shapes have been adjusted in order to map
        # from d_model to num_heads * self.d_head (see below why the projection matrices are not splitted into head-specific matrices).
        # Considering the final output dimensions the W_O matrix will project num_heads*d_head dimensional vectors into d_model vectors, so the whole operation
        # continues to be mathematically consistent ensuring input and output dimension of this module are the same.
        # eg. d_model = 50, num_heads = 4, d_head = int(d_model/num_heads) = 12
        # q = x * W_Q   q shape is 48 (same goes to k and v), W_Q shape is 50x48   (in this example q is the concatention of the q vectors )
        # output = attention_output * W_O   output shape is 50, W_O shape is 48x50
        # if d_model % num_heads != 0:
        #    raise DimError(d_model, num_heads)

        self.d_head = int(d_model / num_heads)  # Query, key and value embeddings dimension. d_k = d_v = d_head

        # Learnable projection matrices. Bias term is omitted since they are used as projections matrices.
        # Every head should have its projection matrix, but rather considering a set of QKV matrices for each head,
        # here 3 bigger matrices are considered. The following example involves the query projection matrix W_Q but the reasoning applies to all of them.
        # Considering D = d_model, d_k = d_v = d_head and A = num_heads.
        # W_Q is a DxD matrix and each W_Q_i (query projection matrix for i-th head) should be a DxD_k matrix.
        # W_Q can be reshaped as a DxAxD_k matrix since A*d_k = D due to initial assertion. (in practice the output projection will be reshaped as mentioned)
        # This way we can properly take advantage of GPU parallelization thanks to torch broadcasting,
        # instead of executing one projection operation at a time for each head in a loop.
        self.W_Q = nn.Linear(d_model, num_heads * self.d_head, bias=False)
        self.W_K = nn.Linear(d_model, num_heads * self.d_head, bias=False)
        self.W_V = nn.Linear(d_model, num_heads * self.d_head, bias=False)
        self.W_O = nn.Linear(num_heads * self.d_head, d_model, bias=False)  # Output projection

        self.dropout = nn.Dropout(dropout_prob)

        self.scaling_factor = math.sqrt(self.d_head)  # To avoid computing it every time attention method is called

    def forward(
        self,
        x_query: Float[torch.Tensor, "B S D"],
        x_key: Float[torch.Tensor, "B S D"],
        x_value: Float[torch.Tensor, "B S D"],
        attention_mask: Bool[torch.Tensor, "B 1 S S"] | None = None,
    ) -> Float[torch.Tensor, "B S D"]:
        """MultiHead attention.

        Args:
            x_query (Float[torch.Tensor, "B S D"]): Matrix of input query embeddings. B is the batch size, S is the sequence length and D is d_model.
            x_key (Float[torch.Tensor, "B S D"]): Matrix of input key embeddings.
            x_value (Float[torch.Tensor, "B S D"]): Matrix of input value embeddings.
            attention_mask (Bool[torch.Tensor, "B 1 S S"] | None, optional): Attention mask to avoid computing attention to padding tokens. It's also used to apply causal masking in decoder self attention. Defaults to None.

        Returns:
            Float[torch.Tensor, "B S D"]: Processed output tensor.
        """
        batch_size = x_query.shape[0]

        # W_Q(x)          [B, S, D]
        # After reshape   [B, S, A, d_k]
        # After transpose [B, A, S, d_k] where A is num_heads, d_k is the head dimension
        query_matrices = self.W_Q(x_query).reshape(batch_size, -1, self.num_heads, self.d_head).transpose(1, 2)
        key_matrices = self.W_K(x_key).reshape(batch_size, -1, self.num_heads, self.d_head).transpose(1, 2)
        value_matrices = self.W_V(x_value).reshape(batch_size, -1, self.num_heads, self.d_head).transpose(1, 2)

        # Concatenated heads outputs
        attn_output = self.attention(query_matrices, key_matrices, value_matrices, attention_mask=attention_mask)

        # Reshape to [B, S, A*d_k]
        attn_output_reshaped = attn_output.transpose(1, 2).reshape(batch_size, -1, self.num_heads * self.d_head)

        # Attention scores, shape [B, S, D]. Combine heads outputs into a single D-dimensional output.
        return self.W_O(attn_output_reshaped)

    def attention(
        self,
        query: Float[torch.Tensor, "B A S d_k"],
        key: Float[torch.Tensor, "B A S d_k"],
        value: Float[torch.Tensor, "B A S d_k"],
        attention_mask: Bool[torch.Tensor, "B 1 S S"] | None = None,
    ) -> Float[torch.Tensor, "B A S d_k"]:
        """Implements attention as follows:

        $$
        Attention(Q,K,V) = softmax ( \\frac{QK^T}{\\sqrt{d_{head}}} ) * V
        $$

        Args:
            query (Float[torch.Tensor, "B, A, S, d_k"]): Matrix of input query embeddings. Where B is the batch size, A is the number of heads, S is the sequence length and d_k is the head dimension.
            key (Float[torch.Tensor, "B, A, S, d_k"]): Matrix of input key embeddings.
            value (Float[torch.Tensor, "B, A, S, d_k"]): Matrix of input value embeddings.
            attention_mask (Bool[torch.Tensor, "B 1 S S"] | None, optional): Attention mask to avoid computing attention to padding tokens. It's also used to apply causal masking in decoder self attention. Defaults to None.

        Returns:
            Float[torch.Tensor, "B, A, S, d_k"]: Attention matrix.
        """
        QKt = torch.matmul(query, key.transpose(-1, -2)) / self.scaling_factor

        if attention_mask is not None:
            # NOTE Adding this control to correctly process masking considering that target input sequence will be shrinked by one token
            # This is especially needed when computing cross attention in decoder blocks due to the usage of src_mask which cannot be shrinked accordingly a priori
            if attention_mask.shape[-1] > QKt.shape[-1] or attention_mask.shape[-2] > QKt.shape[-2]:
                attention_mask = attention_mask[:, :, : QKt.shape[-2], : QKt.shape[-1]]
            QKt.masked_fill_(attention_mask == False, float("-inf"))

        # Applying the softmax on last dim makes results in a QKt matrix with normalized rows
        QKt_norm = torch.softmax(QKt, dim=-1)

        # Store attention weights for visualization purposes
        self.attn_weights = QKt_norm.detach()

        QKt_norm = self.dropout(QKt_norm)

        # Fix nan propagation due to softmax processing of masked matrix containing entire rows full of -inf
        QKt_norm = QKt_norm.masked_fill(torch.isnan(QKt_norm), 0.0)

        return torch.matmul(QKt_norm, value)


class FeedForward(nn.Module):
    def __init__(self, d_model: int, d_ff: int, dropout_prob: float = 0.1):
        super().__init__()
        self.mlp = nn.Sequential(
            nn.Linear(d_model, d_ff),
            nn.SiLU(),
            nn.Dropout(dropout_prob),
            nn.Linear(d_ff, d_model),
        )

    def forward(self, x: Float[torch.Tensor, "B S D"]) -> Float[torch.Tensor, "B S D"]:
        # Pointwise feedforward. It applies the transformation independently and identically to each position along the sequence dimension S and batch size B.
        # This works because nn.Linear operates only on the last dimension D and broadcasts over all leading dimensions,
        # so the same weights are applied to each of the B*S vectors of size D (d_model).
        return self.mlp(x)


class LayerNorm(nn.Module):
    def __init__(self, d_model: int, eps: float = 1e-6):
        super().__init__()
        self.gamma = nn.Parameter(torch.ones(d_model))  # scale
        self.beta = nn.Parameter(torch.zeros(d_model))  # shift
        self.eps = eps  # Avoids ZeroDivisionError

    def forward(self, x: Float[torch.Tensor, "B S D"]) -> Float[torch.Tensor, "B S D"]:
        mean = torch.mean(x, dim=-1, keepdim=True)
        # There's no interest in applying Bessel's correction.
        # It is usually done to make sure var is an unbiased estimator of the sample variance,
        # but here the main goal is to make the output have sum of squares equal to 1
        var = torch.var(x, dim=-1, correction=0, keepdim=True)

        z = (x - mean) / torch.sqrt(var + self.eps)
        return z * self.gamma + self.beta


class EncoderBlock(nn.Module):
    """Transformer Encoder block.

    Args:
        d_model (int): Model dimension.
        num_heads (int): Number of attention heads.
        d_ff (int): Size of middle feedforward layer.
        dropout_prob (float, optional): Dropout probability. Defaults to 0.1.
        norm_type (str, optional): Layer normalization strategy. Defaults to "postnorm".
    """

    def __init__(self, d_model: int, num_heads: int, d_ff: int, dropout_prob: float = 0.1, norm_type: str = "postnorm"):
        super().__init__()

        self.self_attention = MultiHeadAttention(d_model, num_heads, dropout_prob)
        self.feedforward = FeedForward(d_model, d_ff, dropout_prob)
        self.layer_norm1 = LayerNorm(d_model)
        self.layer_norm2 = LayerNorm(d_model)
        # Layer norm at the end of the block for prenorm configuration
        if norm_type == "prenorm":
            self.layer_norm3 = LayerNorm(d_model)
        self.dropout = nn.Dropout(dropout_prob)
        self.norm_type = norm_type

    def forward(
        self, x: Float[torch.Tensor, "B S D"], attention_mask: Bool[torch.Tensor, "B 1 S S"]
    ) -> Float[torch.Tensor, "B S D"]:
        if self.norm_type == "postnorm":
            return self.postnorm_forward(x, attention_mask)
        return self.prenorm_forward(x, attention_mask)

    def postnorm_forward(
        self, x: Float[torch.Tensor, "B S D"], attention_mask: Bool[torch.Tensor, "B 1 S S"]
    ) -> Float[torch.Tensor, "B S D"]:
        """Original Postnorm forward function.

        Tip:
            Accoding to this [paper](https://arxiv.org/abs/2305.09312) it outperforms Prenorm in zero-shot Machine Translation.
        """
        t1 = self.self_attention(x_query=x, x_key=x, x_value=x, attention_mask=attention_mask)
        # We apply dropout to the output of each sub-layer, before it is added to the sub-layer input and normalized (Attention is all you need page 8)
        t2 = self.dropout(t1)
        t3 = t2 + x
        t4 = self.layer_norm1(t3)

        t5 = self.feedforward(t4)
        t6 = self.dropout(t5)
        t7 = t6 + t4
        h = self.layer_norm2(t7)

        return h

    def prenorm_forward(
        self, x: Float[torch.Tensor, "B S D"], attention_mask: Bool[torch.Tensor, "B 1 S S"]
    ) -> Float[torch.Tensor, "B S D"]:
        """Prenorm forward function. Learn more [here](https://arxiv.org/abs/2002.04745)."""
        t1 = self.layer_norm1(x)
        t2 = self.self_attention(x_query=t1, x_key=t1, x_value=t1, attention_mask=attention_mask)
        t3 = self.dropout(t2)
        t4 = t3 + x

        t5 = self.layer_norm2(t4)
        t6 = self.feedforward(t5)
        t7 = self.dropout(t6)
        t8 = t7 + t4

        h = self.layer_norm3(t8)

        return h


class DecoderBlock(nn.Module):
    """Transformer Decoder block.

    Args:
        d_model (int): Model dimension.
        num_heads (int): Number of attention heads.
        d_ff (int): Size of middle feedforward layer.
        dropout_prob (float, optional): Dropout probability. Defaults to 0.1.
        norm_type (str, optional): Layer normalization strategy. Defaults to "postnorm".
    """

    def __init__(self, d_model: int, num_heads: int, d_ff: int, dropout_prob: float = 0.1, norm_type: str = "postnorm"):
        super().__init__()

        self.self_attention = MultiHeadAttention(d_model, num_heads, dropout_prob)
        self.cross_attention = MultiHeadAttention(d_model, num_heads, dropout_prob)
        self.feedforward = FeedForward(d_model, d_ff, dropout_prob)
        self.layer_norm1 = LayerNorm(d_model)
        self.layer_norm2 = LayerNorm(d_model)
        self.layer_norm3 = LayerNorm(d_model)
        # Layer norm at the end of the block for prenorm configuration
        if norm_type == "prenorm":
            self.layer_norm4 = LayerNorm(d_model)
        self.dropout = nn.Dropout(dropout_prob)
        self.norm_type = norm_type

    def forward(
        self,
        x: Float[torch.Tensor, "B S D"],
        encoder_representation: Float[torch.Tensor, "B S D"],
        tgt_mask: Bool[torch.Tensor, "B 1 S S"],
        src_mask: Bool[torch.Tensor, "B 1 S S"],
    ) -> Float[torch.Tensor, "B S D"]:
        if self.norm_type == "postnorm":
            return self.postnorm_forward(x, encoder_representation, tgt_mask, src_mask)
        return self.prenorm_forward(x, encoder_representation, tgt_mask, src_mask)

    def postnorm_forward(
        self,
        x: Float[torch.Tensor, "B S D"],
        encoder_representation: Float[torch.Tensor, "B S D"],
        tgt_mask: Bool[torch.Tensor, "B 1 S S"],
        src_mask: Bool[torch.Tensor, "B 1 S S"],
    ) -> torch.Tensor:
        """Original Postnorm forward function.

        Tip:
            Accoding to this [paper](https://arxiv.org/abs/2305.09312) it outperforms Prenorm in zero-shot Machine Translation.
        """
        t1 = self.self_attention(x_query=x, x_key=x, x_value=x, attention_mask=tgt_mask)
        # We apply dropout to the output of each sub-layer, before it is added to the sub-layer input and normalized (Attention is all you need page 8)
        t2 = self.dropout(t1)
        t3 = t2 + x
        t4 = self.layer_norm1(t3)

        t5 = self.cross_attention(
            x_query=t4, x_key=encoder_representation, x_value=encoder_representation, attention_mask=src_mask
        )

        t6 = self.dropout(t5)
        t7 = t6 + t4
        t8 = self.layer_norm2(t7)

        t9 = self.feedforward(t8)
        t10 = self.dropout(t9)
        t11 = t10 + t8
        h = self.layer_norm3(t11)

        return h

    def prenorm_forward(
        self,
        x: Float[torch.Tensor, "B S D"],
        encoder_representation: Float[torch.Tensor, "B S D"],
        tgt_mask: Bool[torch.Tensor, "B 1 S S"],
        src_mask: Bool[torch.Tensor, "B 1 S S"],
    ) -> torch.Tensor:
        """Prenorm forward function. Learn more [here](https://arxiv.org/abs/2002.04745)."""
        t1 = self.layer_norm1(x)
        t2 = self.self_attention(x_query=t1, x_key=t1, x_value=t1, attention_mask=tgt_mask)
        t3 = self.dropout(t2)
        t4 = t3 + x

        t5 = self.layer_norm2(t4)
        t6 = self.cross_attention(
            x_query=t5, x_key=encoder_representation, x_value=encoder_representation, attention_mask=src_mask
        )
        t7 = self.dropout(t6)
        t8 = t7 + t4

        t9 = self.layer_norm3(t8)
        t10 = self.feedforward(t9)
        t11 = self.dropout(t10)
        t12 = t11 + t8

        h = self.layer_norm4(t12)

        return h


class Transformer(nn.Module):
    """Transformer model.

    Using Language Model head to map decoder output representation to tokens in vocabulary.

    Args:
        src_vocab_size (int): Size of source language vocabulary.
        tgt_vocab_size (int): Size of target language vocabulary.
        num_encoder_blocks (int, optional): Number of encoder blocks. Defaults to 6.
        num_decoder_blocks (int, optional): Number of decoder blocks. Defaults to 6.
        d_model (int, optional): Model dimension. Defaults to 512.
        num_heads (int, optional): Number of heads in MultiHead Attention. Defaults to 8.
        d_ff (int, optional): Size of middle feedforward layer. Defaults to 2048.
        norm_type (str, optional): Layer normalization strategy. Defaults to "postnorm".
        dropout_prob (float, optional): Dropout probability. Defaults to 0.1.
        max_seq_len (int, optional): Max sequence length. Defaults to 128.

    Raises:
        MissingArgumentsError: Raised when `src_emb_from_pretrained` is supplied in `kwargs`, but `src_emb_pretrained_type` and `src_emb_pretrained_path` are not supplied. This error also applies for `tgt_emb_from_pretrained`.
        MissingArgumentsGloVeError: Raised when GloVe embeddings from pretrained are wanted to be loaded and `src_tokenizer` is not supplied. This error also applies for `tgt_emb_from_pretrained`.
    """

    def __init__(
        self,
        src_vocab_size: int,
        tgt_vocab_size: int,
        num_encoder_blocks: int = 6,
        num_decoder_blocks: int = 6,
        d_model: int = 512,
        num_heads: int = 8,
        d_ff: int = 2048,
        norm_type: str = "postnorm",
        dropout_prob: float = 0.1,
        max_seq_len: int = 128,
        **kwargs,
    ):
        super().__init__()

        # Source embedding init
        if kwargs.get("src_emb_from_pretrained") is not None:
            if "src_emb_pretrained_type" not in kwargs or "src_emb_pretrained_path" not in kwargs:
                raise MissingArgumentsError("src")
            if "src_tokenizer" not in kwargs and kwargs["src_emb_pretrained_type"] == "GloVe":
                raise MissingArgumentsGloVeError("src")
            self.src_embeddings = Embedding(
                src_vocab_size,
                d_model,
                from_pretrained=True,
                pretrained_emb_type=kwargs["src_emb_pretrained_type"],
                pretrained_emb_path=kwargs["src_emb_pretrained_path"],
                tokenizer=kwargs["src_tokenizer"],
            )
        else:
            self.src_embeddings = Embedding(src_vocab_size, d_model)

        # Target embedding init
        if kwargs.get("tgt_emb_from_pretrained") is not None:
            if "tgt_emb_pretrained_type" not in kwargs or "tgt_emb_pretrained_path" not in kwargs:
                raise MissingArgumentsError("tgt")
            if "tgt_tokenizer" not in kwargs and kwargs["tgt_emb_pretrained_type"] == "GloVe":
                raise MissingArgumentsGloVeError("tgt")
            self.tgt_embeddings = Embedding(
                tgt_vocab_size,
                d_model,
                from_pretrained=True,
                pretrained_emb_type=kwargs["tgt_emb_pretrained_type"],
                pretrained_emb_path=kwargs["tgt_emb_pretrained_path"],
                tokenizer=kwargs["tgt_tokenizer"],
            )
        else:
            self.tgt_embeddings = Embedding(tgt_vocab_size, d_model)

        self.src_pos_embeddings = SinusoidalPositionalEncoding(d_model, dropout_prob, max_sequence_length=max_seq_len)
        self.tgt_pos_embeddings = SinusoidalPositionalEncoding(d_model, dropout_prob, max_sequence_length=max_seq_len)

        self.encoder = nn.ModuleList([
            EncoderBlock(d_model, num_heads, d_ff, dropout_prob, norm_type) for _ in range(num_encoder_blocks)
        ])
        self.decoder = nn.ModuleList([
            DecoderBlock(d_model, num_heads, d_ff, dropout_prob, norm_type) for _ in range(num_decoder_blocks)
        ])

        self.unembedding_matrix = nn.Linear(d_model, tgt_vocab_size, bias=False)

    def init_params(self):
        """Xavier initialize uninitialized layers.

        This weight initialization strategy was first introduced [here](https://proceedings.mlr.press/v9/glorot10a.html). It's used to stabilize gradients during training.
        """
        for name, p in self.named_parameters():
            if "embeddings" in name:  # Embeddings are initialized in their init function
                continue
            if p.dim() > 1:
                nn.init.xavier_uniform_(p)

    def forward(
        self,
        src_sequence: Float[torch.Tensor, "B S"],
        tgt_sequence: Float[torch.Tensor, "B S"],
        src_mask: Bool[torch.Tensor, "B S"],
        tgt_mask: Bool[torch.Tensor, "B S"],
    ) -> Float[torch.Tensor, "B S D"]:
        encoder_representation = self.encode(src_sequence, src_mask)
        decoder_output = self.decode(tgt_sequence, encoder_representation, tgt_mask, src_mask)

        return decoder_output

    def encode(
        self, src_sequence: Float[torch.Tensor, "B S"], src_mask: Bool[torch.Tensor, "B S"]
    ) -> Float[torch.Tensor, "B S D"]:
        """Encode method."""
        src_x = self.src_embeddings(src_sequence)
        src_x = self.src_pos_embeddings(src_x)

        # Reshape from [B, S] to [B, 1, 1, S] to properly broadcast attention mask to all QKt matrices
        # Broadcasting doc: https://docs.pytorch.org/docs/stable/notes/broadcasting.html
        src_mask = src_mask.unsqueeze(1).unsqueeze(2)

        # Build attention mask matrix with shape [B, 1, S, S] to properly mask QKt matrix
        # eg. with considering only the last 2 dimensions
        # input = [[ True,  True, False]]
        # output = [[ True,  True, False],
        #           [ True,  True, False],
        #           [False, False, False]]
        src_mask = src_mask.to(torch.float)  # Convert to float to allow transpose operation
        src_mask = torch.matmul(src_mask.transpose(-1, -2), src_mask)
        src_mask = src_mask.to(torch.bool)  # Convert back to boolean mask

        # Store mask for visualization purposes
        self.attention_mask = src_mask

        encoder_representation = src_x
        for i in range(len(self.encoder)):
            encoder_representation = self.encoder[i](encoder_representation, src_mask)

        return encoder_representation

    def decode(
        self,
        tgt_sequence: Float[torch.Tensor, "B S"],
        encoder_representation: Float[torch.Tensor, "B S D"],
        tgt_mask: Bool[torch.Tensor, "B S"],
        src_mask: Bool[torch.Tensor, "B S"],
    ) -> Float[torch.Tensor, "B S D"]:
        """Decode method."""
        tgt_x = self.tgt_embeddings(tgt_sequence)
        tgt_x = self.tgt_pos_embeddings(tgt_x)

        tgt_mask = tgt_mask.unsqueeze(1).unsqueeze(2)
        cross_src_mask = src_mask.unsqueeze(1).unsqueeze(2)

        # Build cross attention mask matrix with shape [B, 1, S, S] to properly mask QKt matrix
        # As done before in the encoder attention, now there's the need avoid attention computation to target pad tokens
        # Please refer to the documentation for more details
        # eg. with considering only the last 2 dimensions
        # src_input = [[ True,  True, True, False]]
        # tgt_input = [[ True,  True, False, False]]
        # output = [[ True,  True, True, False],
        #           [ True,  True, True, False],
        #           [False, False, False, False],
        #           [False, False, False, False]]
        cross_src_mask = cross_src_mask.to(torch.float)  # Convert to float to allow transpose operation
        tgt_mask = tgt_mask.to(torch.float)  # Convert to float to allow transpose operation
        cross_src_mask = torch.matmul(tgt_mask.transpose(-1, -2), cross_src_mask)
        cross_src_mask = cross_src_mask.to(torch.bool)  # Convert back to boolean mask

        # Store mask for visualization purposes
        self.cross_attention_mask = cross_src_mask

        tgt_mask = torch.matmul(tgt_mask.transpose(-1, -2), tgt_mask)
        tgt_mask = tgt_mask.to(torch.bool)

        # Apply causal masking
        # This speeds up computation since only one masked_fill will be applied in each decoder attention module
        causal_mask = (
            torch.triu(torch.ones((tgt_mask.shape[0], 1, tgt_mask.shape[-1], tgt_mask.shape[-1])), diagonal=1) == 0
        ).to(tgt_mask.device)
        tgt_mask = tgt_mask & causal_mask  # Extract intersection between pad_mask and causal mask

        # Store mask for visualization purposes
        self.causal_attention_mask = tgt_mask

        decoder_representation = tgt_x
        for i in range(len(self.decoder)):
            decoder_representation = self.decoder[i](
                decoder_representation, encoder_representation, tgt_mask, cross_src_mask
            )

        # Language modeling head
        # Decoded output represents the output logits, softmax needs to be applied if output probabilities are needed
        decoder_output = self.unembedding_matrix(decoder_representation)

        return decoder_output


def build_model(
    config: DictConfig | ListConfig | str,
    src_tokenizer: BaseTokenizer | None = None,
    tgt_tokenizer: BaseTokenizer | None = None,
    from_pretrained: bool | None = False,
    **kwargs,
) -> Transformer:
    """Build Transformer model according to a config file.

    Args:
        config (DictConfig | ListConfig | str): Project config object or str to load it.
        src_tokenizer (BaseTokenizer | None, optional): Source text tokenizer. Defaults to None.
        tgt_tokenizer (BaseTokenizer | None, optional): Target text tokenizer. Defaults to None.
        from_pretrained (bool | None, optional): Load pretrained weights. Defaults to False.

    Raises:
        ModelSizeNotChoosen: Raised when config doesn't have `chosen_model_size` key.

    Returns:
        Transformer: Initialized Transformer model according to config yaml file and choosen model size.
    """

    # Load config
    if isinstance(config, str):
        from omegaconf import OmegaConf

        if config.startswith(("https://", "http://")):
            from io import StringIO

            if not config.endswith((".yaml", ".yml")):
                Exception(
                    f"File extension not supported: `{config.split('.')[-1]}`. Please load a `.yaml` or `.yml` file."
                )
            try:
                resp = requests.get(config, timeout=10)
                resp.raise_for_status()
                config = OmegaConf.load(StringIO(resp.text))
            except Exception as e:
                raise RuntimeError(f"Failed to download config file from {config}: {e}") from e
        else:
            if os.path.exists(config) and config.endswith((".yaml", ".yml")):
                config = OmegaConf.load(config)
            else:
                raise FileNotFoundError(
                    "Config yaml file not found or wrong file provided. Only `.yaml` and `.yml` are allowed."
                )

    if "chosen_model_size" not in config:
        raise ModelSizeNotChoosen()

    # Load from pretrained
    if from_pretrained:
        if kwargs.get("model_path") is not None:
            model = Transformer(
                src_vocab_size=config.src_tokenizer_vocab_size,
                tgt_vocab_size=config.tgt_tokenizer_vocab_size,
                num_encoder_blocks=config.model_configs[config.chosen_model_size].num_encoder_layers,
                num_decoder_blocks=config.model_configs[config.chosen_model_size].num_decoder_layers,
                d_model=config.model_configs[config.chosen_model_size].d_model,
                num_heads=config.model_configs[config.chosen_model_size].num_heads,
                d_ff=config.model_configs[config.chosen_model_size].d_ff,
                norm_type=config.model_configs[config.chosen_model_size].norm_type,
                dropout_prob=config.model_parameters.dropout,
                max_seq_len=config.tokenizer.max_seq_len,
            )

            model_filepath = None
            if str(kwargs["model_path"]).startswith(("https://", "http://")):
                try:
                    response = requests.get(kwargs["model_path"], timeout=100)
                    response.raise_for_status()

                    if kwargs["model_path"].endswith(".safetensors"):
                        suffix = ".safetensors"
                    elif kwargs["model_path"].endswith(".pt"):
                        suffix = ".pt"
                    elif kwargs["model_path"].endswith(".pth"):
                        suffix = ".pth"
                    else:
                        raise InvalidFileExtensionException(kwargs["model_path"].split(".")[-1])

                    with tempfile.NamedTemporaryFile(delete=False, suffix=suffix) as tmp_file:
                        tmp_file.write(response.content)
                        model_filepath = tmp_file.name

                except Exception as e:
                    raise RuntimeError(f"Failed to download model from {kwargs['model_path']}: {e}") from e
            else:
                model_filepath = kwargs["model_path"]

            # Load the model weights
            try:
                if model_filepath.endswith(".safetensors"):
                    state_dict = load_file(model_filepath)
                else:
                    state_dict = torch.load(model_filepath, map_location="cpu", weights_only=True)
                    first_layer_key = next(iter(state_dict.keys()))
                    if first_layer_key.startswith("_orig_mod."):
                        new_state_dict = {}
                        prefix = "_orig_mod."
                        for k, v in state_dict.items():
                            new_k = k[len(prefix) :] if k.startswith(prefix) else k
                            new_state_dict[new_k] = v
                        state_dict = new_state_dict

                model.load_state_dict(state_dict, strict=False, assign=True)

                if str(kwargs["model_path"]).startswith(("https://", "http://")):
                    os.unlink(model_filepath)
            except Exception as e:
                raise RuntimeError(f"Failed to load model weights from {model_filepath}: {e}") from e

            return model

        else:
            raise MissingArgumentsError("Missing model_path argument")

    # Disable pretrained_word_embeddings option when no GloVe embeddings version and filename are provided in the configuration
    if (
        config.model_configs[config.chosen_model_size].get("glove_version", None) is None
        and config.model_configs[config.chosen_model_size].get("glove_filename", None) is None
    ):
        config.model_configs.pretrained_word_embeddings = None

    # NOTE GloVe embeddings available for English only
    if (
        config.dataset.src_lang != "en" and config.dataset.tgt_lang != "en"
    ) or config.model_configs.pretrained_word_embeddings is None:
        model = Transformer(
            src_vocab_size=src_tokenizer.vocab_size,
            tgt_vocab_size=tgt_tokenizer.vocab_size,
            num_encoder_blocks=config.model_configs[config.chosen_model_size].num_encoder_layers,
            num_decoder_blocks=config.model_configs[config.chosen_model_size].num_decoder_layers,
            d_model=config.model_configs[config.chosen_model_size].d_model,
            num_heads=config.model_configs[config.chosen_model_size].num_heads,
            d_ff=config.model_configs[config.chosen_model_size].d_ff,
            norm_type=config.model_configs[config.chosen_model_size].norm_type,
            dropout_prob=config.model_parameters.dropout,
            max_seq_len=config.tokenizer.max_seq_len,
        )
        model.init_params()

    else:
        glove_version = config.model_configs[config.chosen_model_size].glove_version
        glove_filename = config.model_configs[config.chosen_model_size].glove_filename

        glove_path = os.path.join(config.base_path, f"data/{glove_version}/{glove_filename}.txt")

        if config.dataset.src_lang == "en":  # English is the source language
            model = Transformer(
                src_vocab_size=src_tokenizer.vocab_size,
                tgt_vocab_size=tgt_tokenizer.vocab_size,
                num_encoder_blocks=config.model_configs[config.chosen_model_size].num_encoder_layers,
                num_decoder_blocks=config.model_configs[config.chosen_model_size].num_decoder_layers,
                d_model=config.model_configs[config.chosen_model_size].d_model,
                num_heads=config.model_configs[config.chosen_model_size].num_heads,
                d_ff=config.model_configs[config.chosen_model_size].d_ff,
                norm_type=config.model_configs[config.chosen_model_size].norm_type,
                dropout_prob=config.model_parameters.dropout,
                max_seq_len=config.tokenizer.max_seq_len,
                src_emb_from_pretrained=True,
                src_emb_pretrained_type=config.model_configs.pretrained_word_embeddings,
                src_emb_pretrained_path=glove_path,
                src_tokenizer=src_tokenizer,
            )

        else:  # English is the target language
            model = Transformer(
                src_vocab_size=src_tokenizer.vocab_size,
                tgt_vocab_size=tgt_tokenizer.vocab_size,
                num_encoder_blocks=config.model_configs[config.chosen_model_size].num_encoder_layers,
                num_decoder_blocks=config.model_configs[config.chosen_model_size].num_decoder_layers,
                d_model=config.model_configs[config.chosen_model_size].d_model,
                num_heads=config.model_configs[config.chosen_model_size].num_heads,
                d_ff=config.model_configs[config.chosen_model_size].d_ff,
                norm_type=config.model_configs[config.chosen_model_size].norm_type,
                dropout_prob=config.model_parameters.dropout,
                max_seq_len=config.tokenizer.max_seq_len,
                tgt_emb_from_pretrained=True,
                tgt_emb_pretrained_type=config.model_configs.pretrained_word_embeddings,
                tgt_emb_pretrained_path=glove_path,
                tgt_tokenizer=tgt_tokenizer,
            )

        model.init_params()

    return model
